/*
*	XBee tool for Plan9
*	coded by Kenar
*	ver.1.0a
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include "common.h"

typedef struct atcmd atcmd;
struct atcmd {
	char *name;
	int paramsize;
};

/* AT commands list of ZigBee
*	0: read only
*	n (!=0): required n byte in writing	*/
struct atcmd atcmds[] = {
"DH",4,
"DL",4,
"MY",0,
"MP",0,
"NC",0,
"SH",0,
"SL",0,
"NI",20,
"SE",1,
"DE",1,
"CI",1,
"NP",2,
"DD",0,

/* DigiMesh */
"AG",8,
"CE",1,
"CK",1,

"CH",0,
"ID",8,
"MT",1, /* DM */
"OP",8,
"NH",1, /* DM */
"NN",1, /* DM */
"MR",1, /* DM */
"BH",1,
"OI",2,
"NT",1,
"NO",1,
"SC",2,
"SD",1,
"ZS",1,
"NJ",1,
"JV",1,
"NW",1,
"JN",1,
"AR",1,

/* DigiMesh */
"FN",0,

"EE",1,
"EO",1,
"NK",16, /* Network Encryption Key */
"KY",16, /* ink Key */
"PL",1,
"RR",1, /* DM */
"ED",1, /* DM */
"PM",1,
"DB",1,
"PP",1,

/* DM diagnostics */
"BC",0,
"GD",0,
"EA",0,
"TR",0,
"UA",0,
"%H",0,
"%8",0,


"AP",1,
"AO",1,
"BD",1,
"NB",1,
"SB",1,
"RO",1,
"D7",1,
"D6",1,

"IR",2,
"IC",2,
"P0",1,
"P1",1,
"P2",1,
"P3",1,
"D0",1,
"D1",1,
"D2",1,
"D3",1,
"D4",1,
"D5",1,
"D8",1,
"LT",1,
"PR",2,
"M0",2,
"M1",2,
"RP",1,
"%V",2,
"V+",0,
"TP",0,

"VR",0,
"HV",0,
"AI",0,

"CT",2,
"CN",0,
"GT",2,
"CC",1,
"SM",1,
"SN",2,
"SP",2,
"ST",2,
"SO",1,
"WH",2,
"PO",2,
"SI",0,

/* DigiMesh */
"SS",1,
"OS",1,
"OW",1,
"MS",0,
"SQ",1,


"AC",0,
"WR",0,
"RE",0,
"FR",0,
"NR",1,
"SI",0,
"CB",0,
"ND",0,
"DN",0,
"IS",0,
"%V",2,

/* DigiMesh */
"VL",0,

nil,0};


int mid = 0; /* message id (frame id) */
char *dest = nil;
char destbuf[32];
char *marker = "*";
char *init = nil;
int pflag = 1;
int tflag = 0;

#define DBG message
#define LF 10
#define CR 13

#define OFF 0
#define ON	1

void
usage(void)
{
	message("usage: api [-pt] $device\n"
	"	-p: no prompt\n"
	"	-t: show time"
	);
	exits(nil);
}

void
strtoupper(char *s)
{
    while(*s){
        *s = toupper(*s);
        s++;
    }
}

char *
skip(char *s, char *p)
{
	while(*s){
		if(!strchr(p,*s))
			break;
		s++;
	}
	return s;
}

char *
skipto(char *s, char *p)
{
	while(*s){
		if(strchr(p,*s))
			break;
		s++;
	}
	return s;
}

char*
trim(char *s)
{
	char *t;
	while(isspace(*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(isspace(*t ))
		t--;
	t++;
	*t = 0;
	return s;
}

int
hex2bin(char *bp, int m, char *s)
/* we asume that
*	(1) s is hex sequency
*	(2) line ends with NULL
*	example:
*	hex2bin(bp, 3, "a12") then
*	bp[0] = 0
*	bp[1] = 0xa
*	bp[2] = 0x12
*	return value is 2
* NOTE: s will be scratched. Why? This code should be improved	*/
{
	int n,i,k;
	char *endptr;                                          
	for(i=0; i < m; i++)
		bp[i] = 0;
	k = m;
	i = strlen(s);
	while(i > 0 && k > 0){
		s[i] = 0;
		n = i -2;
		if(n < 0)
			n = 0;
		bp[--k] = strtol(&s[n], &endptr,16);
        //debug("k=%d n=%d bp=%02x",k,n,0xff&bp[k]);
		i = n;
	}
	return m - k;
}

int
getpsize(char *name)
{
	atcmd *atp;
	for(atp = atcmds; atp->name; atp++){
		if(strcmp(atp->name,name) == 0)
			break;
	}
	if(atp->name)
		return atp->paramsize;
	return -1;
}

int
checksum(char *s, int n)
{
	int i,sum;
	sum = 0;
	for(i = 0; i < n; i++)
		sum += 0xFF&s[i];
	return 0xFF - (0xFF&sum);
}

int
readln(int fd, char *bp, int n)
/*  return:
 *  -2: error
 *  -1: EOF
 *  others: line length
 */
{
	int k, m;
	for(m = 0; m < n; m++){
		k = read(fd, bp, 1);
		if(k <= 0) /* -1: error or 0: eof */
			return k-1;
		if(*bp == '\n'){
			*bp = 0;
			break;
		}
		bp++;
	}
	return m;
}

void
dump(char *buf,int n)
{
	int i;
	if(n == 0){
		print("\n");
		return;
	}
	for(i = 0; i < n-1; i++)
		print("%02X ",0xFF&buf[i]);
	print("%02X\n",0xFF&buf[n-1]);
}

/* 	convert address such as 0013A200:40A74607 to 0013A20040A74607
   	usage: dest = dpack(dbuf);
	all ':' in dbuf will be removed and the result is returned
	dbuf will be scratched
*/
char *
dpack(char *s)
{
	char *d=s;
	char *d0=s;
	while(*s){
		if(*s == ':')
			s++;
		*d++ = *s++;
	}
	*d = 0;
	return d0;
}

/* frame types */
#define localAT 0x08
#define remoteAT 0x17
#define sendMSG 0x10

/*	make apiframe from line and dest.
	usage: n = pack(buf, sozeof(buf), line, dest)
	the api frame will be stored in buf
	"dest" is the destination address (addr64) or nil
	"line" takes two forms:
	(a)	xx param
	(b)	"msg
*/
int
pack(char *buf, int m, char *line, char *dest)
{
	char dbuf[8];
	char pbuf[256];
	char atname[3];
	char *arg,*p;
	int paramsize=0;
	int maxparamsize;
	int i,n;
	int frametype=remoteAT;
	char *msg = nil;

	if(dest){
		//print("DBG3: %s\n",dest);
		strncpy(pbuf,dest,sizeof(pbuf));
		dest = dpack(pbuf); /* then dest="xxxxxxxxxxxxxxxx" */
		//print("DBG4: %s\n",dest);
		hex2bin(dbuf,8,dest);
		//dump(dbuf,8);//DBG
		dest = dbuf;
	}
	/* now, dest is nil or pointer to addr64 */

	/* the line format is
	 * (a)   XX param
     * where XX is two HD(hexa-decimal letters) and param is one of
	 *    HD string
	 * 	  or
	 *	  NI name
	 *	  (name is a printable string that may include spaces)
	 * or
     * (b)  "mesege
	 *	  (mesege is a printable string that may include spaces)
	 */

	if(line[0] == '"'){
		line++;
		msg = trim(line);
		frametype=sendMSG;
		//print("DBG msg=%s\n",msg);
	}
	else{
		strncpy(atname,line,2);
	    atname[2] = 0;
		maxparamsize = getpsize(atname);
		if(maxparamsize == -1){
			message("%s not a ZigBee command\n",atname);
			return -1;
		}
		arg = skip(&line[2]," \t"); /* skip to the argument */
	    p = skipto(arg," \t"); /* discard comment part */
	    if(*p)
	        *p = 0;
	    
	    if(strcmp(atname,"NI") == 0){
	        strcpy(pbuf,&arg[1]);
	        paramsize = strlen(pbuf);
	        if(paramsize > maxparamsize)
	            paramsize = maxparamsize;
	    }
	    else{
	        paramsize = (strlen(arg)+1)/2; /* 0 -> 0, 1 -> 1, 2 -> 1, 3 -> 2 ... */
	        /* "paramsize == 0" means no argument */
			/* something wrong! */
	        if(paramsize > sizeof(pbuf))
	            paramsize = sizeof(pbuf);
	        if(paramsize != 0)
	            paramsize = maxparamsize;
	        //debug("paramsize=%d",paramsize);
	        hex2bin(pbuf,paramsize,arg);
	    }
	}
        
	/* the XBee data format for localAT request:
	*   1: 7E
	*   2: data size (MSB and LSB)
	*   1: frame type (08 for localAT command)
	*   1: frame ID
	*   2: AT command name
	*	n: params (n depends on command)
	*   1: check sum
	*  the XBee data format for remoteAT request:
	*   1: 7E
	*   2: data size (MSB and LSB)
	*   1: frame type (17 for remoteAT command)
	*   1: frame ID
	*   8: remote address
	*   2: 16 bits destination address (we set FF FE)
	*   1: apply changes (option) # 02 for immediate change
	*   2: AT command name
	*	n: params (n depends on command)
	*   1: check sum
	*  the XBee data format to sendMSG request:
	*   1: 7E
	*   2: data size (MSB and LSB)
	*   1: frame type (10 for sendMSG command)
	*   1: frame ID
	*   8: remote address
	*   2: 16 bits destination address (we set FF FE)
	*   1: broadcast radius (we set 02)
	*   1: option. (00 is enough)
	*	n: RF data (the data)
	*   1: check sum
	*
	*	three types:
	*	ATxx ....		# for remoteAT request
	*	!ATxx ....		# for localAT reques
	*	"....			# for sendMSG request
	*	
	*/
	if(m < 8+paramsize){
		message("buffer size %d too short\n",m);
		return -1;
	}
	buf[0] = 0x7e; /* starting Delimiter */
	n = 3;
	// buf[n++] = dest?0x17:0x08; /* AT command */
	buf[n++] = dest?frametype:localAT;
	/* mid is message ID */
	if(mid == 255)
		mid = 0;
	buf[n++] = ++mid;
	if(dest){
		if(m < 19+paramsize){
			message("buffer size %d too short\n",m);
			return -1;
		}
		for(i = 0; i < 8; i++)
			buf[n++] = dest[i];
		buf[n++] = 0xFF;
		buf[n++] = 0xFE;
		buf[n++] = 0x02;
		/* immediate change for remoteAT
		 * broadcast radius (2 is enough) for sendMSG */
	}
	/* AT command name */
    if(msg){
		buf[n++] = 0x00; /* option */
		//print("DBG: %d %d %s\n",m,n,msg);
		while(*msg)
			buf[n++] = *msg++;
	}
	else{
		buf[n++] = line[0];
		buf[n++] = line[1];
		for(i=0; i < paramsize; i++)
			buf[n++] = pbuf[i];
	}
	/* then data length is n - 3 */
	buf[1] = (n-3)>>8;
	buf[2] = 0xFF&(n-3);
	buf[n] = checksum(&buf[3],n-3);
	n++;
	return n;
}

void
prompt(void)
{
	if(pflag)
		write(2,"> ",2);
}

void
cons2dev(int fd0, int fd1)
{
	char *line, buf[256],pbuf[256];
	int n;

	loop: prompt();
	next:
		/* prompt here doesn't work cleanly. so, I discarded */
		n = readln(fd0, buf, sizeof(buf));
		if(n == -2)
			sysfatal("read error");
        if(n <= 0){ /* eof */
			fprint(2,"use QUIT to exit\n");
			goto loop;; /* it's better not to return */
		}
        line = trim(buf);
        if(line[0] == '#'){
            goto loop;;
		}
		if(line[0] != '"')
        	strtoupper(line);
        if(strcmp(line,"QUIT") == 0)
            return;
		/* then the line is following form:
		 *	 DEST option		# set or look value dest
		 *	 ATxx option		# AT command (may be to local or remote)
		 *	 !ATxx option		# localAT command
		 *	 "message			# sendMSG command (API)
		 * the option is HD(hexa decimal) string except ATNI.
		 * for ATNI, the line is
		 *	 ATNI name
		 * syntax:
		 *	 message,name: consists ASCII printable chars including spaces
		 *   spaces at the beginning will be stripped (not coded yet. required?)
		 */
		if(strcmp(line,"DEST") == 0){
			if(dest)
				print("%s\n",dest);
			else
				print("local\n");
			goto loop;;
		}
		if(strncmp(line,"DEST ",5) == 0){
			char *d;
			d = skip(line+4," ");
			if(strcmp(d,"LOCAL") == 0) // don't replace by local
				dest = nil;
			else{
				/* here, "d" is (for example) "0013A200:40A74509" */
				strncpy(destbuf,d, sizeof(destbuf));
				dest = destbuf;
			}
			goto loop;
		}
		
		if(line[0] == '!'){
			line++;
	        if(strncmp(line,"AT",2) != 0){
            	message("command must begin with AT");
            	goto loop;
			}
			n = pack(pbuf,sizeof(pbuf),line+2,nil);
		}
		else{
			/* then the line is one of two
		 	 *	 ATxx option		# AT command (may be to local or remote)
			 *	 "message			# sendMSG command (API)
			 */	  
	        if(line[0] != '"' && strncmp(line,"AT",2) != 0){
            	message("command must begin with AT");
            	goto loop;
			}
			if(strncmp(line,"AT",2) == 0)
				line += 2;
			n = pack(pbuf,sizeof(pbuf),line,dest);
		}
		if(n < 0)
			goto loop;
		dump(pbuf,n);
		if(write(fd1,pbuf,n) !=n)
			message("error: not be written");
		goto next;
}

void
dev2cons(int fd)
/* read data from the device and print it to console */
{
	char buf[256], *bp;
	char buf0[256];
	int n,len;

	/* we assume: we use consolefs from bell-labs
	 * the device is on /mnt/consoles */
	n = readln(fd,buf,sizeof buf);
	if( n <= 0)
		sysfatal("no data");
	buf[n] = 0;
	/* regular case of buf is "[+arisawa]", otherwise we show it.
     * this message is for preventing multiple access to the device */
	snprint(buf0,sizeof(buf0),"[+%s]",getuser());
	if(strcmp(buf,buf0) != 0){
		fprint(2,"%s\n",buf);
		sysfatal("busy");
	}

	while(1){
		n = read(fd,buf,1);
		if( n <= 0)
			sysfatal("no data");
        /*  1: delimiter
            2: length
            1: frame type
                0x97: remote command response
                0x92: data sample Rx indicator
            1: framid  # not exist for frame type 0x92
            8: 64bits address
            2: 16bits address
            ...
        */
		if(buf[0] != 0x7e){
			print("%02X,",0xff&buf[0]);
			continue;
		}
        if(marker)
            print("%s ",marker);
        if(tflag)
            print("%ld ",time(nil));
		print("%02X ",0xFF&buf[0]); /* 7E */
		n = readn(fd,buf,2); /* MSB */
		if(n != 2)
			sysfatal("no data");
		print("%02X %02X ",0xFF&buf[0], 0xFF&buf[1]);
		bp = buf;
		len = 256*(0xFF&buf[0]) + (0xFF&buf[1]);
		if(len >= 256)
			bp = malloc(len);
        n = readn(fd,bp,len + 1);/* need check sum */
		if( n <= 0)
			sysfatal("no data");
		dump(bp,n);
		if(bp != buf)
			free(bp);
		prompt();
	}
}


void
main(int argc, char *argv[])
{
	int fd;
	int devctl = 0;
    int baud = 0;
	char cname[128];
	int pid;
//	static char dbuf[8];

	ARGBEGIN{
	case 'b':
		baud = atoi(EARGF(usage()));
		break;
	case 'i':
		init = EARGF(usage());
		break;
	case 'p':
		pflag = !pflag;
		break;
	case 't':
		tflag++;
		break;
	default:
		usage();
	}ARGEND

	if(argc != 1)
			usage();

	//message("Opening %s",argv[0]);
	fd = open(argv[0],ORDWR);
	if(fd < 0){
		message("%s not open: %r",argv[0]);
		exits("open");
	}
	//message("%s OK",argv[0]);

	if(baud){
		snprint(cname,sizeof(cname),"%sctl",argv[0]);
		devctl = open(cname,ORDWR);
		if(devctl < 0)
			sysfatal("api: cannot open %sctl: %r\n", cname);
		fprint(devctl, "b%d i1", baud);
	}

    if(init){
        int fd0;
		message("Executing %s",init);
        fd0 = open(init,ORDWR);
        if(fd0 < 0)
            sysfatal("%s not open\n",init);
        cons2dev(fd0,fd);
    }

	switch((pid = fork())){
    case -1: /* error */
        sysfatal("cann't fork");
        break;
    case 0: /* child */
        dev2cons(fd);
        break;
    default: /* parent */
        cons2dev(0,fd);
        kill(pid);
        break;
    }
    if(devctl)
		fprint(devctl, "b%d i1", 9600);
	exits(nil);
}


