#!/bin/9lua
-- api.lua ver.1.1
usage ="usage: api.lua [-p] $device"
-- -p: no prompt (asynchronous)

byte = string.byte
substr = string.sub
char = string.char
fmt = string.format
find = string.find
upper = string.upper
match = string.match
gmatch = string.gmatch
gsub = string.gsub
remove = table.remove
concat = table.concat

prompt = "> "

--[[AT commands list of ZigBee
*	0: read only
*	n (!=0): required n byte in writing	*/

use grep and awk to import table from api.c
	grep '^\"[A-Z]' api.c|awk -F ',' '{print "["$1"]="$2","}'
--]]

atcmd = {
["DH"]=4,
["DL"]=4,
["MY"]=0,
["MP"]=0,
["NC"]=0,
["SH"]=0,
["SL"]=0,
["NI"]=20,
["SE"]=1,
["DE"]=1,
["CI"]=1,
["NP"]=2,
["DD"]=0,
["AG"]=8,
["CE"]=1,
["CK"]=1,
["CH"]=0,
["ID"]=8,
["MT"]=1,
["OP"]=8,
["NH"]=1,
["NN"]=1,
["MR"]=1,
["BH"]=1,
["OI"]=2,
["NT"]=1,
["NO"]=1,
["SC"]=2,
["SD"]=1,
["ZS"]=1,
["NJ"]=1,
["JV"]=1,
["NW"]=1,
["JN"]=1,
["AR"]=1,
["FN"]=0,
["EE"]=1,
["EO"]=1,
["NK"]=16,
["KY"]=16,
["PL"]=1,
["RR"]=1,
["ED"]=1,
["PM"]=1,
["DB"]=1,
["PP"]=1,
["BC"]=0,
["GD"]=0,
["EA"]=0,
["TR"]=0,
["UA"]=0,
["AP"]=1,
["AO"]=1,
["BD"]=1,
["NB"]=1,
["SB"]=1,
["RO"]=1,
["D7"]=1,
["D6"]=1,
["IR"]=2,
["IC"]=2,
["P0"]=1,
["P1"]=1,
["P2"]=1,
["P3"]=1,
["D0"]=1,
["D1"]=1,
["D2"]=1,
["D3"]=1,
["D4"]=1,
["D5"]=1,
["D8"]=1,
["LT"]=1,
["PR"]=2,
["M0"]=2,
["M1"]=2,
["RP"]=1,
["V+"]=2,
["TP"]=0,
["VR"]=0,
["HV"]=0,
["AI"]=0,
["CT"]=2,
["CN"]=0,
["GT"]=2,
["CC"]=1,
["SM"]=1,
["SN"]=2,
["SP"]=2,
["ST"]=2,
["SO"]=1,
["WH"]=2,
["PO"]=2,
["SI"]=0,
["SS"]=1,
["OS"]=1,
["OW"]=1,
["MS"]=0,
["SQ"]=1,
["AC"]=0,
["WR"]=0,
["RE"]=0,
["FR"]=0,
["NR"]=1,
["SI"]=0,
["CB"]=0,
["ND"]=0,
["DN"]=0,
["IS"]=0,
["VL"]=0,
["%V"]=2,
}
-- or
-- loadfile("atcmd2.dat")() -- external atcmd table, if you want

function ppid(pid)
	local t
	s = fmt("/proc/%d/ppid",pid)
	fd = p9.open(s)
	if fd then
		t = p9.read(fd)
		p9.close(fd)
	end
	return t
end

function kill(pid,note)
	local s,fd
	if note == nil then
		note = "note"
	end
	s = fmt("/proc/%d/"..note,pid)
	fd = p9.open(s,"w")
	if fd then
		n = p9.write(fd,"kill")
		p9.close(fd)
	end
	if n ~= 4 then
		print("kill failed")
	end
end

function fatal(...)
	args = {...}
	msg = concat(args,"\t",1,#args)
	print(msg)
	if pid and pid > 0 then
		kill(pid)
	end
	os.exit()
end

function panic(...)
	args = {...}
	msg = concat(args,"\t",1,#args)
	if pid and pid > 0 then
		kill(pid)
	end
	assert(nil,msg)
end


function message(msg)
	if msg ~= prompt then
		msg = msg .. "\n"
	end
	p9.write(2,msg)
end

function trim(s)
	return match(s,"^%s*(.-)%s*$")
end

push = table.insert
function split(s)
	local t,w
	t = {}
	for w in gmatch(s,"%s*(%S+)") do
		push(t,w)
	end
	return t
end

function iscons()
	-- look stat(2) for understanding the code below
	t1=p9.dirstat("/dev/cons")
	t2 = p9.dirstat(0)
	t = {"qid","type","dev"}
	for k=1,3 do
		v = t[k]
		--print(t1[v],t2[v])
		if t1[v] ~= t2[v] then
			return false
		end
	end
	return true
end

function write(fd,msg)
	-- fd ~= 2
	if fd == 1 then
		msg = msg .. "\n"
	end
	return p9.write(fd,msg)
end

function read(fd,n)
	if n == nil then
		s = p9.read(fd)
	else
		s = p9.read(fd,n)
	end
	return s
end


-- read with alarm timeout
-- read n byte
function readn(fd,n)
	local s,t
	--p9.alarm(3)
	if n == nil then
		s = p9.read(fd)
	else
		s = ""
		while #s < n do
			t = p9.read(fd,n - #s)
			if t == nil then
				return nil
			end
			s = s .. t
		end
	end
	--p9.alarm(0)
	return s
end

--[[
the readln() here has two problems
* cannot handle streaming input
* cannot handle mutiple input
look readln.lua
--]]
rdbuf = ""
function readln(fd,opt)
	local s,n,m
	--[[
	opt controles '\n' code at the end of line that is read by readln()
	opt == nil	--> '\n' at the end is discarded frome the line
	opt ~= nil	-->	the line keeps '\n' at the end
	readln(fd) simulates Lua's readln()
	--]]
	m = 1
	if opt then
		m = 0
	end
	if rdbuf == "" then
		rdbuf = p9.read(fd) -- don't use readn(fd) here.
		--print("DBG rdbuf",rdbuf)
		if rdbuf == "" then
			return nil
		end
	end
	--print("DBG rdbuf",rdbuf)
	n = find(rdbuf,"\n", 1, true)
	if n == nil then
		-- this happens only in regular file
		s = substr(rdbuf, 1, #rdbuf)
		rdbuf = ""
		return s
	end
	s = substr(rdbuf, 1, n-m)
	rdbuf = substr(rdbuf, n+1, #rdbuf)
	--print("DBG",rdbuf)
	return s
end

function hpack(s)
	-- print("-- hpack",s)
	local t
	s = split(s) 	-- then s is a table of string
	t = ""
	for i=1, #s do
		u = s[i]
		if #u%2 == 1 then
			u = "0" .. u
		end
		t = t .. u
	end
	--print(t)
	return t
end
--[[
hpack("7E 00 0F 17") --> "7E000F17"
hpack("7E 0 0F 117") --> "7E000F0117"
hpack("7E00 0F 117") --> "7E000F0117"
hpack("7E000F0117") --> "7E000F0117"
os.exit()
--]]

--[[
let's hstr denotes space separated bytes series such as:
7E 00 0F 17 03 00 13 A2 00 40 A7 45 09 FF FE 02 49 53 60
hex2str converts hstr to a string of this internal code
--]]

--[[
convert string such that
0013A20040A74509
to a string of this internal code
--]]
function hex2str(s)
	-- print("-- hex2str",s)
	local t,b,u
	t = hpack(s)
	b = ""
	for i=1,#t//2 do
		u = substr(t,2*i-1,2*i)
		b = b .. char(tonumber(u,16))
	end
	--print(b)
	return b
end
--[[
hex2str("61 62 63")	--> "abc"
hex2str("6162 63")	--> "abc"
hex2str("616263")	--> "abc"
os.exit()
--]]


-- str2hex is a converse to hex2str
function str2hex(s)
	local t
	--print("-- str2hex",#s,s)
	t = ""
	for i=1, #s-1 do
		t = t..fmt("%02X ",byte(s,i))
	end
	return t..fmt("%02X",byte(s,#s))
end
--[[
h=str2hex("abc")
print(h)	--> 61 62 63
print(hex2str(h))	--> abc
h=str2hex("")
print(#h, h)	--> 0
print(#hex2str(h))	--> 0
print(str2hex(hex2str("0013A20040A74509")))	--> 00 13 A2 00 40 A7 45 09
os.exit()
--]]

--[[
we need a function that reads frame data of XBee.
rdframe() read frame from fd and check the validity
return the frame
--]]
function rdframe(fd)
	--print("--rdframe fd",fd)
	if pid and pid > 0 then
		panic("# parent process must not read device")
	end
	::loop::
	s = readn(fd,1)
	if s == nil then
		return nil
	end
	--print("DBG s",s)
	if substr(s,1,1) ~= "~" then
		message("# invalid frame header: " .. s .. " " .. str2hex(s))
		-- read until "~"
		message("# trying to get next frame")
		goto loop
--[[
		while s and s ~= "~" do
			s = readn(fd,1)
		end
--]]
	end
	s = s .. readn(fd,2)
	t = {byte(s,1,3)}	-- then t is an array of integer
	len = t[2]*256 + t[3]	-- body length
	s = s .. readn(fd,len+1) -- including last chksum
	sum = mksum(substr(s,4,#s))
	if sum ~= 255 then
		message("# check sum error: " .. str2hex(s))
		return nil
	end
	return s
end

--[[
mksum() is used to check the checksum
and also used to make checksum
--]]
function mksum(s)
	local t
	t = {byte(s,1,#s)}
	sum = 0
	for i=1,#t do
		sum = sum + t[i]
	end
	return (0xff & sum)
end

--[[
-- I illustrate here how check sum is computed
function hprint(x)
	print(fmt("%02X",x))
end

-- a frame data s:
s="7E 00 0D 88 02 49 44 00 00 00 00 00 00 00 00 01 E7"

-- in the frame data we have:
t="88 02 49 44 00 00 00 00 00 00 00 00 01"
u=mksum(hex2str(t))
hprint(0xff-u)	--> should be "E7"

t="88 02 49 44 00 00 00 00 00 00 00 00 01 E7"
u=mksum(hex2str(t))
hprint(u)	--> should be "255"

os.exit()
--]]



--[[ the XBee data format for localAT request:
*   1: 7E
*   2: data size (MSB and LSB)
*   1: frame type (08 for localAT command)
*   1: frame ID
*   2: AT command name
*	n: params (n depends on command)
*   1: check sum
*  the XBee data format for remoteAT request:
*   1: 7E
*   2: data size (MSB and LSB)
*   1: frame type (17 for remoteAT command)
*   1: frame ID
*   8: remote address
*   2: 16 bits destination address (we set FF FE)
*   1: apply changes (option) # 02 for immediate change
*   2: AT command name
*	n: params (n depends on command)
*   1: check sum
*  the XBee data format to sendMSG request:
*   1: 7E
*   2: data size (MSB and LSB)
*   1: frame type (10 for sendMSG command)
*   1: frame ID
*   8: remote address
*   2: 16 bits destination address (we set FF FE)
*   1: broadcast radius (we set 02)
*   1: option. (00 is enough)
*	n: RF data (the data)
*   1: check sum
*
*	three type:
*	ATxx ....		# for remoteAT request
*	!ATxx ....		# for localAT reques
*	"....			# for sendMSG request
*	
--]]

-- global constants:
localAT=char(0x08)
remoteAT=char(0x17)
sendMSG=char(0x10)		-- Faludi p.131

function mkframe(line,dest)
	-- dest is nil or 8B address
	--print("-- mkframe",line,str2hex(line))
	local remote, param,msg,maxparamsize,ftype,data
	local fid1,sum,fbody,bodysize,frame

	remote = ""
	param = ""
	if dest then
		--param = dest -- need exmination
		dest = gsub(dest,":","")
		remote = hex2str(dest)
	end
	if substr(line,1,1) == '"' then
		msg = substr(line,2,#line)
		if not dest then
			message("# need destination")
			return nil
		end
	else
		cmd = substr(line,1,2)
		cmd = upper(cmd)
		maxparamsize = atcmd[cmd]
		if maxparamsize == nil then
			message("# not a ZigBee command: "..cmd)
			return nil
		end
		line = substr(line,3,#line)
		if cmd == "NI" then
			param = trim(line)
		else
			line = upper(line)
			line = gsub(line,"[ ]","")
			--print("DBG2",line)
			param = hex2str(line)		-- line is given by hstr
			--assert(#param <= atcmd[cmd], "# param size exceeds limit:"..line)
			if #param > atcmd[cmd] then
				message("# param size exceeds limit")
				return nil
			end
		end
	end

	addr16 = char(0xff,0xfe)

	if dest then
		if msg then
			ftype = sendMSG
			data = remote .. addr16 .. char(0x02,0x00) .. msg	-- p.131
		else
			ftype = remoteAT
			data = remote .. addr16 .. char(0x02) .. cmd .. param -- p.143
		end
	else
		ftype = localAT
		data = cmd .. param		-- p.126
	end

	fid1 = char(fid)

	fbody = ftype .. fid1 .. data

	sum = char(0xff - mksum(fbody))
	bodysize = char(#fbody >> 8, 0xff & #fbody)
	frame = "~" .. bodysize .. fbody .. sum
	return frame
end
--[[ some example here
frame=mkframe("id",nil)
print(str2hex(frame))	--> 7E 00 04 08 02 49 44 68
frame=mkframe("d05",nil)
print(str2hex(frame))	--> 7E 00 05 08 02 44 30 05 7C
frame=mkframe("D04",nil)
print(str2hex(frame))	--> 7E 00 05 08 03 44 30 04 7C
frame=mkframe("SL","0013A200:40A74509")
print(str2hex(frame))	--> 7E 00 0F 17 01 00 13 A2 00 40 A7 45 09 FF FE 02 53 4C 5F
os.exit()

-- spaces in AT command are removed
-- frame=mkframe("d0 5",nil) should be also OK
--]]

dest = nil
fid = 0

-- need global fid
function incfid()
	fid = (fid+1) %256
	if fid == 0 then
		fid = 1;
	end
end

--[[
convert cmd line style format to the function style format.
the example:
"aa bb cc"	-> "aa(bb,cc)"
--]]
function line2func(s)
	-- print("-- line2func",s)
	local t,u
	t = split(s)
	if #t == 1 then
		return t[1] .. "()"
	end
	u = t[1] .. "(" .. t[2]
	for i=3,#t do
		u = u .. "," .. t[i]
	end
	u = u .. ")"
	return u
end

function test1(fd)
	-- try: "test1 fd" in command mode
	-- Transmit (TX) Request: IPv4 - 0x20
	-- look doc/s6b.txt for details
	local data,frame,chksum,len
	data = char(0x20,0x01)..char(192,168,0,6)..char(0x26,0x16)..char(0x26,0x16)..char(0,0).."Hello"
	len = char(#data//256,#data%256)
	chksum = 0xff - mksum(data)
	frame = char(0x7E)..len..data..char(chksum)
	write(1,str2hex(frame))
	assert(write(fd,frame) == #frame, "# write error")
	return true
end
--[[
hebe# api6.lua /mnt/consoles/usb14
test1 fd
7E 00 11 20 01 C0 A8 00 06 26 16 26 16 00 00 48 65 6C 6C 6F 04
* 7E 00 03 89 01 00 75
--]]




-- console to device
function cons2dev(fd)
	--print("DBG -- cons2dev",fd)
	local line, head, args, cmd, t, ntry, maxtry
	-- we have two types of errors:
	-- (a) coding error -- exit with error message
	-- (b) input error  -- allow the user to retry
	maxtry = 3;
	::loop::
	ntry = 0
	::retry::
	if ntry > maxtry then
		message("the error count exceeded limit")
		return false
	end
	ntry = ntry + 1
	if prompt then
		message(prompt)
	end
	line = readln(0)
	if line == nil then os.exit() end	-- for EOF
	if #line == 0 then goto loop end
	-- the line exclude "\n" at the end. therefore #line == 0 for empty line 
	line = trim(line)
	head = substr(line,1,1)
	--print("DBG:",#head,head)
	if head == "#" then
		goto loop
	end
	incfid()
	if head == '"' then
		frame = mkframe(line,dest)
		if frame == nil then
			goto retry
		end
		write(1,str2hex(frame))
		assert(write(fd,frame) == #frame,"# write error")
		--goto loop
		if prompt then
			return true
		else
			goto loop
		end
	end
	if head == '!' then
		-- local request
		line = match(line,"! *(.*)")
		--print("DBG",#line,line)
		frame = mkframe(substr(line,3,#line),nil)
		write(1,str2hex(frame))
		assert(write(fd,frame) == #frame,"# write error")
		goto loop
	end

	args = split(line)

	-- line = match(line,"[ \t]*(.*)")
	-- cmd = match(line,"([_%w]+)")
	cmd = args[1]
	--print("DBG cmd:",cmd)
	if cmd == nil then
		message("no such command")
		goto retry
	end

	-- print("DBG cmd",cmd)
	cmdu = upper(cmd)
	if cmdu == "QUIT" then
		return false
	end
	if cmdu == "LOAD" then 
		t = loadfile(args[2])
		if t == nil then
			message("# syntax error in "..args[2])
			goto retry
		end
		t()
		goto loop
	end
	if cmdu == "DEST" then
		if args[2] then
			dest = upper(args[2])
			if dest == "LOCAL" then
				dest = nil
			end
			goto loop
		end
		if dest then
			write(1,dest)
		else
			write(1,"local")
		end
		goto loop
	end
	if substr(cmdu,1,2) == "AT" then
		frame = mkframe(substr(line,3,#line),dest)
		if frame == nil then
			goto loop
		end
		write(1,str2hex(frame))
		assert(write(fd,frame) == #frame, "# write error")
		--print("DBG write dev OK")
		if prompt then
			return true
		else
			goto loop
		end
	end

	if _G[cmd] then
		--print("DBG _G")
		f = line2func(line)
		--print("DBG f",f)
		r = load(f)()	-- r is nil, not reurn value
		-- try: r = load(return f)()
		--print("DBG load result:",r)
		goto loop
	end

	message(cmd.." not found")
	goto retry

end

-- device to console
function dev2cons(fd)
	--print("DBG -- dev2cons",fd)
	local s,n
	n = 0
	::loop::
	s = rdframe(fd)
	if s == nil then
		if n > 3 then
			kill(pid0)
			fatal("# rdframe")
		end
		n = n + 1
		goto loop
	end
	write(1,"* "..str2hex(s))
	if not prompt then
		goto loop
	end
end

-- read from console
fd = nil
pid = nil
pid0 = p9.getpid()
pflag = nil
function cons()
	local s,t,user,ms
	-- the input of api.lua may be a pipe
	if iscons() == false then
		prompt = nil
	end
	if arg[1] == "-p" then
		pflag = true
		prompt = nil
		remove(arg,1)
	end
	if #arg ~= 1 then
		fatal(usage)
	end
	fd = p9.open(arg[1],"rw")
	if fd == nil then
		fatal("not open "..arg[1])
	end
	t = "/mnt/consoles/"
	if substr(arg[1],1,#t) ~= t then
		fatal("# Invalid argument", arg[1])
	end

	s = trim(read(fd))
	--print("DBG s",s);
	user = os.getenv("user")
	ms = match(s,"%[%+(.*)%]")
	if ms ~= user then
		fatal("busy. someone is using:",ms)
	end

	if pflag then
		NOTEG = 1<<3; p9.fork(NOTEG)
		-- this makes easy to clean up KB-interruption
		-- look rfork(2) and /sys/include/libc.h for NOTEG
		pid = p9.fork()
		if pid > 0 then -- parent
			cons2dev(fd);
			kill(pid)
		elseif pid == 0 then  -- child
			dev2cons(fd)
			os.exit()
		else
	        fatal("cann't fork");
		end
	else
		while cons2dev(fd) do
			p9.alarm(3)
			dev2cons(fd)
			p9.alarm(0)
		end
	end
end

cons()

--[[
need API mode
use att.lua and
	ATAP	-- look current mode
	ATAP 1	-- API-enabled
--]]


