/*
* att for Plan9
* code by Kenar
*/

#include <u.h>
#include <libc.h>
#include "common.h"

#define EOF 4
#define DEL 127
#define LF 10
#define CR 13

#define OFF 0
#define ON	1

void
usage(void)
{
	message("usage: att [-q] [-b baud] $device\n");
	exits(nil);
}

void
constodev(int fd)
/* read chars from the console and send it to modem */
{
	char buf[1];
	while(1){
		if(read(0,buf,1) != 1)
			break;
		if(buf[0] == EOF || buf[0] == DEL)
			break;
		if(write(1,buf,1) != 1)
			sysfatal("error wrie:%r");
		if(buf[0] == LF)
			buf[0] = CR;
		if(write(fd,buf,1) != 1)
			sysfatal("error wrie:%r");
		/* write(fd,buf,1) != 1 does not mean
		 * the fd is connected to xbee
		 * for example: vid 0x067b did 0x2303 does not make error
		 * even if it is disconnected from xbee */
	}
}

void
devtocons(int fd)
/* read chars from the modem and send it to console */
{
	char buf[1];
	while(1){
		if(read(fd,buf,1) != 1)
			sysfatal("error read:%r");
		if(buf[0] == CR)
			buf[0] = LF;
		if(write(1,buf,1) != 1)
			sysfatal("error wrie:%r");
		/* no respose from vid 0x067b did 0x2303
		 * Prolific Technology Inc.USB-Serial Controller */
	}
}


void
main(int argc, char *argv[])
{
	int fd,consctl;
	int devctl = 0;
    int baud = 0;
	char cname[128];
	int pid;

	ARGBEGIN{
	case 'b':
		baud = atoi(EARGF(usage()));
		break;
	default:
		usage();
	}ARGEND

	if(argc != 1)
        usage();

	message("Opening %s",argv[0]);
	fd = open(argv[0],ORDWR);
	if(fd < 0){
		message("%s not open",argv[0]);
		exits("open");
	}

	consctl = open("/dev/consctl", OWRITE);
	if(consctl < 0)
		sysfatal("att: cannot open /dev/consctl: %r\n");

	if(baud){
		snprint(cname,sizeof(cname),"%sctl",argv[0]);
		devctl = open(cname,ORDWR);
		if(devctl < 0)
			sysfatal("att: cannot open %sctl: %r\n", cname);
		fprint(devctl, "b%d i1", baud);
	}

	switch((pid = fork())){
    case -1: /* error */
        sysfatal("cann't fork");
        break;
	case 0: /* child */
		devtocons(fd);
		exits(nil);
	default: /* parent */
		raw(consctl,ON);
		constodev(fd);
		kill(pid);
 		raw(consctl,OFF);
	}
    if(devctl)
		fprint(devctl, "b%d i1", 9600);

	exits(nil);
}
/* att doesn't work for API mode.
 * for XBee S2, use API tools such as api.lua or xbee/api instead.
 * under raw(consctl,ON) keybord input comes to console inmediatly.
 * you will find example /usr/arisawa/doc/cons/a.c
 * that show the effect of raw on/off */